//! Challenges and their common interface.
//!
//! This module is the core of the Wooly Vault application, as it defines the interface that all
//! challenges must implement, and contains the challenge modules themselves.

use anyhow::anyhow;
use async_trait::async_trait;

use crate::config::Config;
use crate::vault::VaultRef;

pub mod c1;
pub mod c2;
pub mod c3;

/// Defines the behavior of a challenge.
///
/// Any type that implements this trait can be served as a challenge in the Wooly Vault application.
#[async_trait]
pub trait Challenge
where
    Self: Sized,
{
    /// Creates a new instance of the challenge with the given configuration and vault.
    ///
    /// # Arguments
    ///
    /// * `config` - The configuration for the challenge.
    /// * `vault` - The vault that holds the secret for the challenge.
    ///
    /// # Returns
    ///
    /// A new instance of the challenge.
    fn new(config: Config, vault: VaultRef) -> Self;
    /// Returns a list of hints for the challenge.
    ///
    /// A hint is a short text to be given to the contestants in case the host thinks they need
    /// it. The first hint is the most vague, afterwards the hints become more and more helpful.
    ///
    /// # Returns
    ///
    /// A vector of strings containing hints for the challenge.
    fn hints() -> Vec<String>;
    /// Returns the solution to the challenge.
    ///
    /// A solution is a short description of what has to be done to solve the challenge and get the
    /// secret of the vault.
    ///
    /// # Returns
    ///
    /// A string containing the solution to the challenge.
    fn solution() -> String;
    /// Starts the challenge and serves it to the contestants.
    ///
    /// This method is asynchronous and returns a result indicating whether the challenge was
    /// successful only after the challenge has ended.
    ///
    /// # Returns
    ///
    /// A result indicating whether the challenge was successful ended.
    ///
    /// # Errors
    ///
    /// Will error when the challenge errors, for example when the network adress cannot be bound.
    async fn serve(self) -> anyhow::Result<()>;
}

/// Selects a challenge by index and serves it with the given configuration and vault.
///
/// # Arguments
///
/// * `index` - The index of the challenge to select.
/// * `config` - The configuration for the challenge.
/// * `vault` - The vault that holds the secret for the challenge.
///
/// # Returns
///
/// A result indicating whether the challenge has successfully ended.
///
/// # Errors
///
/// Returns an error if no challenge with the given index exists, or if the challenge that is being
/// served errors.
pub async fn select_and_start(index: u16, config: Config, vault: VaultRef) -> anyhow::Result<()> {
    match index {
        1 => c1::C1::new(config, vault).serve().await?,
        2 => c2::C2::new(config, vault).serve().await?,
        3 => c3::C3::new(config, vault).serve().await?,
        _ => {
            return Err(anyhow!(
                "no challenge with index {index} does currently exist"
            ))
        }
    }
    Ok(())
}
